import asyncio
import logging
import typing as t
from inspect import isawaitable

from typing_extensions import ParamSpec, Self

_T = t.TypeVar("_T")
_P = ParamSpec("_P")
TyImpl = t.Callable[_P, t.Union[t.Awaitable[_T], _T]]

log = logging.getLogger(__name__)


async def call_impls(
    impls: t.Iterable[TyImpl[_P, _T]], *args: _P.args, **kwds: _P.kwargs
) -> t.AsyncGenerator[_T, t.Any]:
    """A uniform interface to call a batch of hook implements."""

    for listener in impls:
        try:
            c = listener(*args, **kwds)
        except:
            log.error("sync listener error!", exc_info=True)
            continue

        if isawaitable(c):
            try:
                c = await c
            except asyncio.CancelledError:
                raise
            except:
                log.error("async listener error!", exc_info=True)
                continue

        yield c  # type: ignore


class StaticHookSpec(t.Generic[_P, _T]):
    """Define a hook. The wrapped function will be the original defination."""

    __slots__ = ("__def__",)

    def __init__(self, defination: TyImpl[_P, _T]) -> None:
        self.__def__ = defination

    def instantiate(self) -> "HookSpec[_P, _T]":
        """Get a :class:`HookSpec` of this hook."""
        return HookSpec(self)

    __call__ = instantiate

    @property
    def TyInst(self) -> "t.Type[HookSpec[_P, _T]]":
        """Designed for type checkers. Used to annotate the type of the instantiated hook."""
        return HookSpec


class HookSpec(t.Generic[_P, _T]):
    """An instance of a :class:`StaticHookSpec`."""

    __slots__ = ("__def__", "impls")

    impls: t.List[TyImpl[_P, _T]]
    """You can access this list to add/remove/clear_all implements."""

    def __init__(self, hookdef: StaticHookSpec[_P, _T]) -> None:
        super().__init__()
        self.impls = []
        self.__def__ = hookdef.__def__

    def add_impl(self, impl: TyImpl[_P, _T]) -> Self:
        """A shortcut to `.impls.append`."""
        self.impls.append(impl)
        return self

    async def gather(self, *args: _P.args, **kwds: _P.kwargs) -> t.List[_T]:
        """Gather all results, results respsect the corresponding order in `.impls`."""
        return [i async for i in call_impls(self.impls, *args, **kwds)]

    results = gather
    """An alias to :meth:`.gather`."""

    async def emit(self, *args: _P.args, **kwds: _P.kwargs) -> None:
        """Like `.gather`, but returns `None`."""
        await self.gather(*args, **kwds)

    async def first(self, *args: _P.args, **kwds: _P.kwargs) -> _T:
        """Get the first valid result. If all results are invalid, raise `StopAsyncIteration`."""
        async for i in call_impls(self.impls, *args, **kwds):
            return i
        raise StopAsyncIteration

    async def __call__(self, *args: _P.args, **kwds: _P.kwargs) -> _T:
        """
        Get the first valid result.
        If all results are invalid, call the defination and return its result.
        """
        try:
            return await self.first(*args, **kwds)
        except StopAsyncIteration:
            c = self.__def__(*args, **kwds)
            return await c if isawaitable(c) else c  # type: ignore

    @property
    def has_impl(self):
        """Return if a hook is implemented."""
        return bool(self.impls)



import asyncio
import typing as t

T = t.TypeVar("T")


class FutureStore:
    """
    A store for keeping references of running futures,
    in order to keep them from being garbage collected.
    """

    _futs: t.Set[asyncio.Future]

    def __init__(self) -> None:
        super().__init__()
        self._futs = set()

    def add_awaitable(self, func: t.Awaitable[T]) -> "asyncio.Future[T]":
        """Add an awaitable into this store.

        :param func: the awaitable
        :return: the wrapped task
        """
        if not (func := asyncio.ensure_future(func)).done():
            self._futs.add(func)
            func.add_done_callback(self._futs.discard)
        return func

    __call__ = add_awaitable

    async def wait(self, wait_new=True):
        """Wait for all tasks in the specific group(s).

        :param wait_new: also wait for new-added futures during waiting for existing futures.

        .. note:: Even if `wait_new` is False, newly added tasks may still be scheduled.
        """
        if not self._futs:
            return
        await asyncio.wait(self._futs)
        if wait_new and self._futs:
            # await potential new tasks in this store
            await self.wait()

    def clear(self):
        """Clear current future store, cancel all saved tasks. If you do not want a task to be cancelled,
        wrap it in a `~asyncio.shield`.

        :param exc: if not given, all futures will be cancelled; else they will be set with this exception.
        """
        for t in self._futs:
            t.cancel()

    def __bool__(self):
        return bool(self._futs)



"""A tiny hook specification library with typing support."""

from typing import Any

from .futstore import FutureStore
from .hookspec import HookSpec, StaticHookSpec

__all__ = [
    "StaticHookSpec",
    "HookSpec",
    "hookdef",
    "null_emitter",
    "FutureStore",
]

hookdef = StaticHookSpec
null_emitter: HookSpec[Any, Any] = StaticHookSpec(lambda *args, **kwds: None).instantiate()
